@@ -1,8 +1,8 @@
 # Module: MachineLearning
 # Author: Adrian Antico <adrianantico@gmail.com>
 # License: MIT
-# Release: retrofit 0.1.5
-# Last modified : 2021-09-20
+# Release: retrofit 0.1.6
+# Last modified : 2021-09-22
 
 def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsList=None, TargetColumnName=None, NumericColumnNames=None, CategoricalColumnNames=None, TextColumnNames=None, WeightColumnName=None, Threads=-1, Processing='catboost', InputFrame='datatable'):
     """
@@ -25,77 +25,6 @@ def ML0_GetModelData(TrainData=None, ValidationData=None, TestData=None, ArgsLis
     Threads:                Number of threads to utilize if available for the algorithm
     Processing:             'catboost', 'xgboost', 'lightgbm', or 'ftrl'
     InputFrame:             'datatable', 'polars', or 'pandas' If you input Frame is 'pandas', it will be converted to a datatable Frame for generating the new columns
-
-    # ML0_GetModelData Example:
-    import pkg_resources
-    import timeit
-    import datatable as dt
-    from datatable import sort, f, by
-    import retrofit
-    from retrofit import FeatureEngineering as fe
-    from retrofit import MachineLearning as ml
-
-    # Load some data
-    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
-    data = dt.fread(FilePath)
-        
-    # Create partitioned data sets
-    DataSets = fe.FE2_AutoDataParition(
-      data=data, 
-      ArgsList=None, 
-      DateColumnName='CalendarDateColumn', 
-      PartitionType='random', 
-      Ratios=[0.70,0.20,0.10], 
-      ByVariables=None, 
-      Processing='datatable', 
-      InputFrame='datatable', 
-      OutputFrame='datatable')
-
-    # Collect partitioned data
-    TrainData = DataSets['TrainData']
-    ValidationData = DataSets['ValidationData']
-    TestData = DataSets['TestData']
-    del DataSets
-
-    # Create catboost data sets
-    t_start = timeit.default_timer()
-    DataSets = ml.ML0_GetModelData(
-      TrainData=TrainData, 
-      ValidationData=ValidationData, 
-      TestData=TestData, 
-      ArgsList=None, 
-      TargetColumnName='Leads', 
-      NumericColumnNames=['XREGS1', 'XREGS2', 'XREGS3'], 
-      CategoricalColumnNames=['MarketingSegments','MarketingSegments2','MarketingSegments3','Label'], 
-      TextColumnNames=None, 
-      WeightColumnName=None, 
-      Threads=-1, 
-      Processing='catboost', 
-      InputFrame='datatable')
-      
-    # timer
-    t_end = timeit.default_timer()
-    t_end - t_start
-    
-    # Collect catboost training data
-    catboost_train = DataSets['train_data']
-    catboost_validation = DataSets['validation_data']
-    catboost_test = DataSets['test_data']
-    ArgsList = DataSets['ArgsList']
-    
-    # QA: Group Case: Step through function
-    TrainData=TrainData
-    ValidationData=ValidationData
-    TestData=TestData
-    ArgsList=None
-    TargetColumnName='Leads'
-    NumericColumnNames=['XREGS1','XREGS2','XREGS3']
-    CategoricalColumnNames=['MarketingSegments', 'MarketingSegments2', 'MarketingSegments3', 'Label']
-    TextColumnNames=None
-    WeightColumnName=None
-    Threads=-1
-    Processing='catboost'
-    InputFrame='datatable'
     """
     
     # For making copies of lists so originals aren't modified
@@ -429,59 +358,6 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None):
     import retrofit
     from retrofit import FeatureEngineering as fe
     from retrofit import MachineLearning as ml
-
-    # Load some data
-    FilePath = pkg_resources.resource_filename('retrofit', 'datasets/BenchmarkData.csv') 
-    data = dt.fread(FilePath)
-        
-    # Create partitioned data sets
-    DataSets = fe.FE2_AutoDataParition(
-      data=data, 
-      ArgsList=None, 
-      DateColumnName='CalendarDateColumn', 
-      PartitionType='random', 
-      Ratios=[0.70,0.20,0.10], 
-      ByVariables=None, 
-      Processing='datatable', 
-      InputFrame='datatable', 
-      OutputFrame='datatable')
-
-    # Collect partitioned data
-    TrainData = DataSets['TrainData']
-    ValidationData = DataSets['ValidationData']
-    TestData = DataSets['TestData']
-    del DataSets
-
-    # Create catboost data sets
-    DataSets = ml.ML0_GetModelData(
-      TrainData=TrainData, 
-      ValidationData=ValidationData, 
-      TestData=TestData, 
-      ArgsList=None, 
-      TargetColumnName='Leads', 
-      NumericColumnNames=['XREGS1', 'XREGS2', 'XREGS3'], 
-      CategoricalColumnNames=['MarketingSegments','MarketingSegments2','MarketingSegments3','Label'], 
-      TextColumnNames=None, 
-      WeightColumnName=None, 
-      Threads=-1, 
-      Processing='catboost', 
-      InputFrame='datatable')
-
-    # Collect Args
-    Args = DataSets.get('ArgsList')
-
-    # Create Parameters for Modeling
-    ModelArgs = ml.ML0_Parameters(
-      Algorithms='catboost', 
-      TargetType='regression', 
-      TrainMethod='Train')
-
-    # QA
-    Algorithms='catboost'
-    TargetType='regression'
-    TrainMethod='Train'
-    Model=None
-    Algo = 'catboost'
     """
     
     # Args Check
@@ -831,7 +707,6 @@ def ML0_Parameters(Algorithms=None, TargetType=None, TrainMethod=None):
     return MasterArgs
 
 
-# RetroFit Class 
 class RetroFit:
     """
     ####################################
@@ -1397,8 +1272,7 @@ class RetroFit:
 
       # Setup Environment
       import datatable as dt
-      from datatable.models import Ftrl
-
+      
       #################################################
       # Ftrl Method
       #################################################
@@ -1406,6 +1280,7 @@ class RetroFit:
 
         # Setup Environment
         from datatable import f
+        from datatable.models import Ftrl
         
         # Extract model
         if not ModelName is None:
@@ -1605,12 +1480,12 @@ class RetroFit:
         else:
           ScoreData = NewData
 
-        # Subset score data columns
-        ScoreData = ScoreData[:, self.DataSets.get('ArgsList').get('NumericColumnNames')]
+        # Subset score features
+        scor = ScoreData[:, self.DataSets.get('ArgsList').get('NumericColumnNames')]
         
         # Regression and Classification
         if TempArgs.get('TargetType').lower() != 'multiclass':
-          ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(data = ScoreData)
+          ScoreData[f"Predict_{TargetColumnName}"] = Model.predict(data = scor)
           
           # Non regression cases
           if TempArgs.get('TargetType').lower() == 'classification':
@@ -1619,7 +1494,7 @@ class RetroFit:
 
         # MultiClass
         else:
-          preds = dt.Frame(Model.predict(data = ScoreData))
+          preds = dt.Frame(Model.predict(data = scor))
           if not self.DataSets.get('ArgsList')['MultiClass'] is None:
             from datatable import cbind
             temp = self.DataSets.get('ArgsList')['MultiClass']
@@ -1645,59 +1520,196 @@ class RetroFit:
     #################################################
     #################################################
     
-    # regression metrics helper
-    def _regression_metrics(self, _FitName = None, y_true = None, y_pred = None):
+    # Evaluation Attribute Update
+    def ML1_Single_Evaluate(self, FitName=None, TargetType=None, ScoredDataName=None, ByVariables=None, CostDict=dict(tpcost = 0.0, fpcost = 1.0, fncost = 1.0, tncost = 0.0)):
+      
+      # TargetType Agnostic Imports
       import datatable as dt
       from datetime import datetime
-      from sklearn.metrics import explained_variance_score, max_error, mean_absolute_error, mean_squared_error, mean_squared_log_error, mean_absolute_percentage_error, median_absolute_error, r2_score
-      Metrics = dt.Frame([self.FitList[_FitName]])
-      Metrics.names = {'C0': 'ModelName'}
-      Metrics['FeatureSet'] = None
-      Metrics['CreateTime'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
-      Metrics['explained_variance_score'] = explained_variance_score(y_true, y_pred)
-      Metrics['r2_score'] = r2_score(y_true, y_pred)
-      Metrics['mean_absolute_percentage_error'] = mean_absolute_percentage_error(y_true, y_pred)
-      Metrics['mean_absolute_error'] = mean_absolute_error(y_true, y_pred)
-      Metrics['median_absolute_error'] = median_absolute_error(y_true, y_pred)
-      Metrics['mean_squared_error'] = mean_squared_error(y_true, y_pred)
-      Metrics['mean_squared_log_error'] = mean_squared_log_error(y_true, y_pred)
-      Metrics['max_error'] = max_error(y_true, y_pred)
-      return(Metrics)
-    
-    # classification metrics helper
-    def _classification_metrics(self, _FitName = None, y_true = None, y_pred = None):
-      import datatable
-      from datetime import datetime
-      from sklearn.metrics import accuracy_score, balanced_accuracy_score
-      Metrics = dt.Frame([self.FitList[_FitName]])
-      Metrics.names = {'C0': 'ModelName'}
-      Metrics['CreateTime'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
-      Metrics['accuracy_score'] = accuracy_score(y_true, y_pred)
-      
-      y_true.to_list()
-      
-      Metrics['balanced_accuracy_score'] = balanced_accuracy_score(y_true, y_pred)
-      Metrics['mean_absolute_percentage_error'] = mean_absolute_percentage_error(y_true, y_pred)
-      Metrics['mean_absolute_error'] = mean_absolute_error(y_true, y_pred)
-      Metrics['median_absolute_error'] = median_absolute_error(y_true, y_pred)
-      Metrics['mean_squared_error'] = mean_squared_error(y_true, y_pred)
-      Metrics['mean_squared_log_error'] = mean_squared_log_error(y_true, y_pred)
-      Metrics['max_error'] = max_error(y_true, y_pred)
-      return(Metrics)
-    
-    # Evaluation Attribute Update
-    def ML1_Single_Evaluate(self, FitName=None, TargetType=None, ScoredDataName=None, ByVariables=None):
-      
+      import numpy as np
+
       # Get Data
+      TargetColumnName = self.DataSets.get('ArgsList').get('TargetColumnName')
       temp = self.DataSets.get(ScoredDataName)
-        
+
       # Generate metrics
-      if TargetType == 'regression':
-        Metrics = _regression_metrics(self, _FitName = FitName, y_true = temp[TargetColumnName], y_pred = temp[f"Predict_{TargetColumnName}"])
-      elif TargetType == 'classification':
-        Metrics = _classification_metrics(self, _FitName = FitName, y_true = temp[TargetColumnName], y_pred = temp[f"Predict_{TargetColumnName}"])
-      elif TargetType == 'multiclass':
-        Metrics = _multiclass_metrics(self, _FitName = FitName, y_true = temp[TargetColumnName], y_pred = temp[f"Predict_{TargetColumnName}"])
-      
-      # Store Metrics
-      self.EvaluationList[f"Eval_{FitName}_{ScoredDataName}"] = Metrics
+      if TargetType.lower() == 'regression':
+
+        # Environment
+        from sklearn.metrics import explained_variance_score, max_error, mean_absolute_error, mean_squared_error, mean_squared_log_error, mean_absolute_percentage_error, median_absolute_error, r2_score
+
+        # Actuals and preds
+        y_true = temp[TargetColumnName]
+        y_pred = temp[f"Predict_{TargetColumnName}"]
+
+        # checks
+        Min_y_true = min(y_true.to_numpy())[0]
+        Min_y_pred = min(y_pred.to_numpy())[0]
+        check = (Min_y_true > 0) & (Min_y_pred > 0)
+
+        # Metrics
+        Metrics = dt.Frame(ModelName = [FitName])
+        Metrics['FeatureSet'] = None
+        Metrics['CreateTime'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
+        if ByVariables:
+          Metrics['Grouping'] = ByVariables
+        else:
+          Metrics['Grouping'] = 'NA'
+        Metrics['explained_variance_score'] = explained_variance_score(y_true, y_pred)
+        Metrics['r2_score'] = r2_score(y_true, y_pred)
+        Metrics['mean_absolute_percentage_error'] = mean_absolute_percentage_error(y_true, y_pred)
+        Metrics['mean_absolute_error'] = mean_absolute_error(y_true, y_pred)
+        Metrics['median_absolute_error'] = median_absolute_error(y_true, y_pred)
+        Metrics['mean_squared_error'] = mean_squared_error(y_true, y_pred)
+        if check:
+          Metrics['mean_squared_log_error'] = mean_squared_log_error(y_true, y_pred) 
+        else:
+          Metrics['mean_squared_log_error'] = -1
+        Metrics['max_error'] = max_error(y_true, y_pred)
+        return Metrics
+
+      # Generate metrics
+      if TargetType.lower() == 'classification':
+
+        # Imports
+        from datatable import ifelse, math, f, update
+
+        # Cost matrix
+        tpcost = CostDict['tpcost']
+        fpcost = CostDict['fpcost']
+        fncost = CostDict['fncost']
+        tncost = CostDict['tncost']
+
+        # Build metrics table
+        Thresholds = list(np.linspace(0.0, 1.0, 101))
+        ThreshLength = [-1.0] * len(Thresholds)
+        ThresholdOutput = dt.Frame(
+          ModelName   = [FitName] * len(Thresholds),
+          FeatureSet  = [None] * len(Thresholds),
+          Grouping    = [ByVariables] * len(Thresholds),
+          CreateTime  = [datetime.now().strftime("%Y-%m-%d %H:%M:%S")] * len(Thresholds),
+          Threshold   = Thresholds,
+          TN          = ThreshLength,
+          TP          = ThreshLength,
+          FN          = ThreshLength,
+          FP          = ThreshLength,
+          N           = ThreshLength,
+          P           = ThreshLength,
+          Utility     = ThreshLength,
+          MCC         = ThreshLength,
+          Accuracy    = ThreshLength,
+          F1_Score    = ThreshLength,
+          F2_Score    = ThreshLength,
+          F0_5_Score  = ThreshLength,
+          TPR         = ThreshLength,
+          TNR         = ThreshLength,
+          FNR         = ThreshLength,
+          FPR         = ThreshLength,
+          FDR         = ThreshLength,
+          FOR         = ThreshLength,
+          NPV         = ThreshLength,
+          PPV         = ThreshLength,
+          ThreatScore = ThreshLength)
+
+        # Generate metrics
+        counter = 0
+        for Thresh in Thresholds:
+          TN = temp[:, dt.sum(ifelse((f['p1'] < Thresh) & (f[TargetColumnName] == 0), 1, 0))].to_list()[0][0]
+          TP = temp[:, dt.sum(ifelse((f['p1'] > Thresh) & (f[TargetColumnName] == 1), 1, 0))].to_list()[0][0]
+          FN = temp[:, dt.sum(ifelse((f['p1'] < Thresh) & (f[TargetColumnName] == 1), 1, 0))].to_list()[0][0]
+          FP = temp[:, dt.sum(ifelse((f['p1'] > Thresh) & (f[TargetColumnName] == 0), 1, 0))].to_list()[0][0]
+          N1 = temp.shape[0]
+          N  = temp[f["p1"] < Thresh, ...].shape[0]
+          P1 = temp[f[TargetColumnName] == 1, ...].shape[0]
+          P  = temp[(f[TargetColumnName] == 1) & (f['p1'] > Thresh), ...].shape[0]
+
+          # Calculate metrics ----
+          if not ((TP+FP) == 0 or (TP+FN) == 0 or (TN+FP) == 0 or (TN+FN) == 0):
+            MCC         = (TP*TN-FP*FN)/np.sqrt((TP+FP)*(TP+FN)*(TN+FP)*(TN+FN))
+          else:
+            MCC = -1.0
+          if not N1 == 0:
+            Accuracy    = (TP+TN)/N1
+          else:
+            Accuracy = -1.0
+          if not P1 == 0:
+            TPR         = TP/P1
+          else:
+            TPR = -1.0
+          if not (N1-P1) == 0:
+            TNR         = TN/(N1-P1)
+          else:
+            TNR = -1.0
+          if not P1 == 0:
+            FNR         = FN / P1
+          else:
+            FNR = -1.0
+          if not N1 == 0:
+            FPR         = FP / N1
+          else:
+            FPR = -1.0
+          if not (FP + TP) == 0:
+            FDR         = FP / (FP + TP)
+          else:
+            FDR = -1.0
+          if not (FN + TN) == 0:
+            FOR         = FN / (FN + TN)
+          else:
+            FOR = -1.0
+          if not (TP + FP + FN) == 0:
+            F1_Score    = 2 * TP / (2 * TP + FP + FN)
+          else:
+            F1_Score = -1.0
+          if not (TP + FP + FN) == 0:
+            F2_Score    = 3 * TP / (2 * TP + FP + FN)
+          else:
+            F2_Score = -1.0
+          if not (TP + FP + FN) == 0:
+            F0_5_Score  = 1.5 * TP / (0.5 * TP + FP + FN)
+          else:
+            F0_5_Score = -1.0
+          if not (TN + FN) == 0:
+            NPV         = TN / (TN + FN)
+          else:
+            NPV = -1.0
+          if not (TP + FP) == 0:
+            PPV         = TP / (TP + FP)
+          else:
+            PPV = -1.0
+          if not (TP + FN + FP) == 0:
+            ThreatScore = TP / (TP + FN + FP)
+          else:
+            ThreatScore = -1.0
+          if not ((N1 == 0) or (TPR == -1.0) or (FPR == -1.0)):
+            Utility     = P1/N1 * (tpcost * TPR + fpcost * (1 - TPR)) + (1 - P1/N1) * (fncost * FPR + tncost * (1 - FPR))
+          else:
+            Utility = -1.0
+
+          # Fill in values ----
+          ThresholdOutput[counter, update(P = P)]
+          ThresholdOutput[counter, update(N = N)]
+          ThresholdOutput[counter, update(TN = TN)]
+          ThresholdOutput[counter, update(TP = TP)]
+          ThresholdOutput[counter, update(FP = FP)]
+          ThresholdOutput[counter, update(FN = FN)]
+          ThresholdOutput[counter, update(Utility = Utility)]
+          ThresholdOutput[counter, update(MCC = MCC)]
+          ThresholdOutput[counter, update(Accuracy = Accuracy)]
+          ThresholdOutput[counter, update(F1_Score = F1_Score)]
+          ThresholdOutput[counter, update(F0_5_Score= F0_5_Score)]
+          ThresholdOutput[counter, update(F2_Score = F2_Score)]
+          ThresholdOutput[counter, update(NPV = NPV)]
+          ThresholdOutput[counter, update(TPR = TPR)]
+          ThresholdOutput[counter, update(TNR = TNR)]
+          ThresholdOutput[counter, update(FNR = FNR)]
+          ThresholdOutput[counter, update(FPR = FPR)]
+          ThresholdOutput[counter, update(FDR = FDR)]
+          ThresholdOutput[counter, update(FOR = FOR)]
+          ThresholdOutput[counter, update(PPV = PPV)]
+          ThresholdOutput[counter, update(ThreatScore = ThreatScore)]
+          
+          # Increment
+          counter = counter + 1
+
+        # return
+        return ThresholdOutput
